@@ -2,21 +2,22 @@
  * 
  * stdio.c: stdio calls
  * 
- * Copyright 2006,2007 Renzo Davoli University of Bologna - Italy
+ * Copyright 2006-2017 Renzo Davoli University of Bologna - Italy
  * 
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * The GNU C Library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License a
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with the GNU C Library; if not, write to the Free
+ * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
  */ 
 
 #include <config.h>
@@ -170,52 +171,6 @@ FILE *fdopen (int fd, const char *modes){
 	return new_pure_file(fd,modes);
 }
 
-int fileno (FILE *stream){
-	/*char buf[]="FNOxx\n";
-		buf[3]=(stream == NULL)?'X':'-';
-		buf[4]=(!_pure_magic(stream))?'X':'-';
-		write (2,buf,6);*/
-	if (stream == NULL) 
-		return -1;
-	else {
-		int rv=0;
-		int (*_fileno)()=dlsym(RTLD_NEXT,"fileno");
-		if (_fileno) {
-			rv=_fileno(stream);
-			if (rv<0) {
-				rv=_pure_file_fd(stream,1);
-				if (rv>=0)
-					errno=0;
-			}
-		} else
-			rv= -1;
-		return rv;
-	}
-}
-
-int fileno_unlocked (FILE *stream){
-	/*char buf[]="FNOxx\n";
-		buf[3]=(stream == NULL)?'X':'-';
-		buf[4]=(!_pure_magic(stream))?'X':'-';
-		write (2,buf,6);*/
-	if (stream == NULL) 
-		return -1;
-	else {
-		int rv=0;
-		int (*_fileno)()=dlsym(RTLD_NEXT,"fileno");
-		if (_fileno) {
-			rv=_fileno(stream);
-			if (rv<0) {
-				rv=_pure_file_fd(stream,0);
-				if (rv>=0)
-					errno=0;
-			}
-		} else
-			rv= -1;
-		return rv;
-	}
-}
-
 FILE *tmpfile (void){
 	int fd;
 	char template[20] = "/tmp/tmpfile-XXXXXX";
@@ -335,3 +290,54 @@ int puts(const char *s)
 		rv=putc('\n',stdout);
 	return rv;
 }
+
+/* this is convenient since casting the return value of dlsym() to
+ * a function pointer erroneously procudes a warning */
+#pragma GCC diagnostic ignored "-Wpedantic"
+
+int fileno (FILE *stream){
+	/*char buf[]="FNOxx\n";
+		buf[3]=(stream == NULL)?'X':'-';
+		buf[4]=(!_pure_magic(stream))?'X':'-';
+		write (2,buf,6);*/
+	if (stream == NULL) 
+		return -1;
+	else {
+		int rv=0;
+		int (*_fileno)()=dlsym(RTLD_NEXT,"fileno");
+		if (_fileno) {
+			rv=_fileno(stream);
+			if (rv<0) {
+				rv=_pure_file_fd(stream,1);
+				if (rv>=0)
+					errno=0;
+			}
+		} else
+			rv= -1;
+		return rv;
+	}
+}
+
+int fileno_unlocked (FILE *stream){
+	/*char buf[]="FNOxx\n";
+		buf[3]=(stream == NULL)?'X':'-';
+		buf[4]=(!_pure_magic(stream))?'X':'-';
+		write (2,buf,6);*/
+	if (stream == NULL) 
+		return -1;
+	else {
+		int rv=0;
+		int (*_fileno)()=dlsym(RTLD_NEXT,"fileno");
+		if (_fileno) {
+			rv=_fileno(stream);
+			if (rv<0) {
+				rv=_pure_file_fd(stream,0);
+				if (rv>=0)
+					errno=0;
+			}
+		} else
+			rv= -1;
+		return rv;
+	}
+}
+
