@@ -1,25 +1,25 @@
 /*   This is part of pure_libc (a project related to ViewOS and Virtual Square)
- *  
+ *
  *   syscalls.c: syscall mgmt
- *   
- *   Copyright 2006-2017 Renzo Davoli University of Bologna - Italy
+ *
+ *   Copyright 2006-2020 Renzo Davoli University of Bologna - Italy
  *   Copyright 2005 Andrea Gasparini University of Bologna - Italy
- *   
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
- * 
+ *
  * The GNU C Library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with the GNU C Library; if not, write to the Free
  * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  * 02110-1301, USA.
- */ 
+ */
 
 #include <config.h>
 #include <stdarg.h>
@@ -57,6 +57,7 @@
 #include <time.h>
 #include <grp.h>
 #include <limits.h>
+#include <sched.h>
 #include <stdlib.h>
 #include "purelibc.h"
 
@@ -64,7 +65,7 @@
 sfun _pure_syscall = syscall;
 sfun _pure_native_syscall;
 
-#if __WORDSIZE != 64
+#if defined(__NR_mmap2)
 static int _pageshift()
 {
 	static int ps=0;
@@ -92,10 +93,23 @@ long _pure_debug_printf(const char *format, ...)
 	return rv;
 }
 
-/* DAMNED! the kernel stat are different! so glibc converts the 
- *  * kernel structure. We have to make the reverse conversion! */
-
 // open must consider two mode of calling: with two or three arguments
+int __open_2(const char* pathname,int flags){
+#if defined(__NR_openat) && ! defined(__NR_open)
+	return _pure_syscall(__NR_openat,AT_FDCWD,pathname,flags);
+#else
+	return _pure_syscall(__NR_open,pathname,flags);
+#endif
+}
+
+static int __open_3(const char* pathname,int flags, mode_t mode) {
+#if defined(__NR_openat) && ! defined(__NR_open)
+	return _pure_syscall(__NR_openat,AT_FDCWD,pathname,flags,mode);
+#else
+	return _pure_syscall(__NR_open,pathname,flags,mode);
+#endif
+}
+
 int open(const char* pathname,int flags,...){
 	va_list arg_list;
 	if( flags &  O_CREAT ){
@@ -103,14 +117,10 @@ int open(const char* pathname,int flags,...){
 		va_start(arg_list,flags);
 		mode = va_arg(arg_list,mode_t);
 		va_end(arg_list);
-		return _pure_syscall(__NR_open,pathname,flags,mode);
+		return __open_3(pathname,flags,mode);
 	}
 	else
-		return _pure_syscall(__NR_open,pathname,flags);
-}
-
-int __open_2(const char* pathname,int flags){
-	return _pure_syscall(__NR_open,pathname,flags);
+		return __open_2(pathname,flags);
 }
 
 int open64(const char* pathname,int flags,...){
@@ -120,24 +130,24 @@ int open64(const char* pathname,int flags,...){
 		va_start(arg_list,flags);
 		mode = va_arg(arg_list,mode_t);
 		va_end(arg_list);
-		return _pure_syscall(__NR_open,pathname,flags|O_LARGEFILE,mode);
+		return __open_3(pathname,flags|O_LARGEFILE,mode);
 	}
 	else
-		return _pure_syscall(__NR_open,pathname,flags|O_LARGEFILE);
+		return __open_2(pathname,flags|O_LARGEFILE);
 }
 
 int __open64_2 (const char* pathname,int flags){
-	return _pure_syscall(__NR_open,pathname,flags|O_LARGEFILE);
+	return __open_2(pathname,flags|O_LARGEFILE);
 }
 
 int creat(const char *pathname, mode_t mode)
 {
-	return _pure_syscall(__NR_open,pathname,O_CREAT|O_WRONLY|O_TRUNC,mode);
+	return __open_3(pathname,O_CREAT|O_WRONLY|O_TRUNC,mode);
 }
 
 int creat64(const char *pathname, mode_t mode)
 {
-	return _pure_syscall(__NR_open,pathname,O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE,mode);
+	return __open_3(pathname,O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE,mode);
 }
 
 int close(int fd){
@@ -169,12 +179,16 @@ int dup(int oldfd){
 }
 
 int dup2(int oldfd, int newfd){
+#if defined(__NR_dup3) && ! defined(__NR_dup2)
+	return _pure_syscall(__NR_dup3, oldfd, newfd, 0);
+#else
 	return _pure_syscall(__NR_dup2, oldfd, newfd);
+#endif
 }
 
 #ifdef __NR_dup3
 int dup3(int oldfd, int newfd, int flags){
-	  return _pure_syscall(__NR_dup3, oldfd, newfd,flags);
+	  return _pure_syscall(__NR_dup3, oldfd, newfd, flags);
 }
 #endif
 
@@ -188,10 +202,23 @@ int dup3(int oldfd, int newfd, int flags){
  * myself. The following not-so-readable stuff takes care of calling the
  * correct 64-bit function on both 32 bit and 64 bit architectures.*/
 
-#if __WORDSIZE == 64
+#ifdef __NR_fstatat64
+#define __NR_FSTATAT64 __NR_fstatat64
+#endif
+#ifdef __NR_newfstatat
+#define __NR_FSTATAT64 __NR_newfstatat
+#endif
+
+#if __WORDSIZE == 64 || defined(__ILP32__)
+# if defined(__NR_FSTATAT64) && ! defined(__NR_stat)
+#  define __USE_FSTATAT64
+# endif
 #	define arch_stat64 stat
 #	define IFNOT64(x)
 #else
+# if defined(__NR_FSTATAT64) && ! defined(__NR_stat64)
+#  define __USE_FSTATAT64
+# endif
 #	define arch_stat64 stat64
 #	define IFNOT64(x) x
 #endif
@@ -199,7 +226,7 @@ int dup3(int oldfd, int newfd, int flags){
 #define INTERNAL_MAKE_NAME(a, b) a ## b
 #define MAKE_NAME(a, b) INTERNAL_MAKE_NAME(a, b)
 
-void arch_stat64_2_stat(struct arch_stat64 *from, struct stat *to)
+static void arch_stat64_2_stat(struct arch_stat64 *from, struct stat *to)
 {
 	if ((void*)from == (void*)to)
 		return;
@@ -229,7 +256,11 @@ int __xstat(int ver, const char* pathname, struct stat* buf_stat)
 	switch(ver)
 	{
 		case _STAT_VER_LINUX:
+#ifdef __USE_FSTATAT64
+			rv = _pure_syscall(__NR_FSTATAT64, AT_FDCWD, pathname, MAKE_NAME(buf_, arch_stat64), 0);
+#else
 			rv = _pure_syscall(MAKE_NAME(__NR_, arch_stat64), pathname, MAKE_NAME(buf_, arch_stat64));
+#endif
 			break;
 
 		default:
@@ -251,7 +282,11 @@ int __lxstat(int ver, const char* pathname, struct stat* buf_stat)
 	switch(ver)
 	{
 		case _STAT_VER_LINUX:
+#ifdef __USE_FSTATAT64
+			rv = _pure_syscall(__NR_FSTATAT64, AT_FDCWD, pathname, MAKE_NAME(buf_, arch_stat64), AT_SYMLINK_NOFOLLOW);
+#else
 			rv = _pure_syscall(MAKE_NAME(__NR_l, arch_stat64), pathname, MAKE_NAME(buf_, arch_stat64));
+#endif
 			break;
 
 		default:
@@ -286,60 +321,90 @@ int __fxstat(int ver, int fildes, struct stat* buf_stat)
 }
 
 int __xstat64(int ver,const char* pathname,struct stat64* buf){
+#ifdef __USE_FSTATAT64
+	return _pure_syscall(__NR_FSTATAT64, AT_FDCWD, pathname, buf, 0);
+#else
 	return _pure_syscall(MAKE_NAME(__NR_, arch_stat64), pathname, buf);
+#endif
 }
 
 int __lxstat64(int ver,const char* pathname,struct stat64* buf){
+#ifdef __USE_FSTATAT64
+	return _pure_syscall(__NR_FSTATAT64, AT_FDCWD, pathname, buf,	AT_SYMLINK_NOFOLLOW);
+#else
 	return _pure_syscall(MAKE_NAME(__NR_l, arch_stat64), pathname, buf);
+#endif
 }
 
 int __fxstat64 (int ver, int fildes, struct stat64 *buf){
 	return _pure_syscall(MAKE_NAME(__NR_f, arch_stat64), fildes, buf);
 }
 /* end of unreadable code */
+#ifdef __NR_statx
+int statx(int dirfd, const char *pathname, int flags,
+		unsigned int mask, struct statx *statxbuf) {
+	return _pure_syscall(__NR_statx, dirfd, pathname, flags, mask, statxbuf);
+}
+#endif
 
 int mknod(const char *pathname, mode_t mode, dev_t dev) {
+#if defined(__NR_mknodat) && ! defined(__NR_mknod)
+	return _pure_syscall(__NR_mknodat,AT_FDCWD,pathname,mode,dev);
+#else
 	return _pure_syscall(__NR_mknod,pathname,mode,dev);
+#endif
 }
 
 int __xmknod (int ver, const char *path, mode_t mode, dev_t *dev) {
+#if defined(__NR_mknodat) && ! defined(__NR_mknod)
+	return _pure_syscall(__NR_mknodat,AT_FDCWD,path,mode,dev);
+#else
 	return _pure_syscall(__NR_mknod,path,mode,dev);
+#endif
 }
 
 int access(const char* pathname,int mode){
+#if defined(__NR_faccessat) && ! defined(__NR_access)
+	return _pure_syscall(__NR_faccessat,AT_FDCWD,pathname,mode,0);
+#else
 	return _pure_syscall(__NR_access,pathname,mode);
+#endif
 }
 
 int __access(const char* pathname,int mode){
 	return access(pathname,mode);
 }
 
-#ifdef __NR_faccessat
-int euidaccess(const char *pathname, int mode){
-	return _pure_syscall(__NR_faccessat,AT_FDCWD,pathname,mode,AT_EACCESS);
-}
-int eaccess(const char *pathname, int mode){
-	return euidaccess(pathname,mode);
-}
-int __euidaccess(const char *pathname, int mode){
-	return euidaccess(pathname,mode);
-}
-#endif
-
 ssize_t readlink(const char* pathname,char* buf, size_t bufsize){
+#if defined(__NR_readlinkat) && ! defined(__NR_readlink)
+	return _pure_syscall(__NR_readlinkat,AT_FDCWD,pathname,buf,bufsize);
+#else
 	return _pure_syscall(__NR_readlink,pathname,buf,bufsize);
+#endif
 }
 
 int mkdir(const char* pathname,mode_t mode){
+#if defined(__NR_mkdirat) && ! defined(__NR_mkdir)
+	return _pure_syscall(__NR_mkdirat,AT_FDCWD,pathname,mode);
+#else
 	return _pure_syscall(__NR_mkdir,pathname,mode);
+#endif
 }
 
 int rmdir(const char* pathname){
+#if defined(__NR_unlinkat) && ! defined(__NR_rmdir)
+	return _pure_syscall(__NR_unlinkat,AT_FDCWD,pathname,AT_REMOVEDIR);
+#else
 	return _pure_syscall(__NR_rmdir,pathname);
+#endif
 }
 
 int chmod(const char* pathname,mode_t mode){
+#if defined(__NR_fchownat) && ! defined(__NR_chmod)
+	return _pure_syscall(__NR_fchownat,AT_FDCWD,pathname,mode,0);
+#else
 	return _pure_syscall(__NR_chmod,pathname,mode);
+#endif
 }
 
 int fchmod(int fd,mode_t mode){
@@ -347,31 +412,55 @@ int fchmod(int fd,mode_t mode){
 }
 
 int chown(const char* pathname,uid_t owner,gid_t group){
+#if defined(__NR_fchownat) && ! defined(__NR_chown)
+	return _pure_syscall(__NR_fchownat,AT_FDCWD,pathname,owner,group,0);
+#else
 	return _pure_syscall(__NR_chown,pathname,owner,group);
+#endif
 }
 
 int lchown(const char* pathname,uid_t owner,gid_t group){
+#if defined(__NR_fchownat) && ! defined(__NR_lchown)
+	return _pure_syscall(__NR_fchownat,AT_FDCWD,pathname,owner,group,AT_SYMLINK_NOFOLLOW);
+#else
 	return _pure_syscall(__NR_lchown,pathname,owner,group);
+#endif
 }
 
 int fchown(int fd,uid_t owner,gid_t group){
 	return _pure_syscall(__NR_fchown,fd,owner,group);
 }
 
-int link(const char* pathname,const char* newpath){
+int link(const char* pathname,const char*newpath){
+#if defined(__NR_linkat) && ! defined(__NR_link)
+	return _pure_syscall(__NR_linkat,AT_FDCWD,pathname,AT_FDCWD,newpath,0);
+#else
 	return _pure_syscall(__NR_link,pathname,newpath);
+#endif
 }
 
 int unlink(const char* pathname){
+#if defined(__NR_unlinkat) && ! defined(__NR_unlink)
+	return _pure_syscall(__NR_unlinkat,AT_FDCWD,pathname,0);
+#else
 	return _pure_syscall(__NR_unlink,pathname);
+#endif
 }
 
 int symlink(const char* pathname,const char* newpath){
+#if defined(__NR_symlinkat) && ! defined(__NR_symlink)
+	return _pure_syscall(__NR_symlinkat,pathname,AT_FDCWD,newpath,0);
+#else
 	return _pure_syscall(__NR_symlink,pathname,newpath);
+#endif
 }
 
 int rename(const char *oldpath, const char *newpath){
+#if defined(__NR_renameat2) && ! defined(__NR_rename)
+	return _pure_syscall(__NR_renameat2,AT_FDCWD,oldpath,AT_FDCWD,newpath,0);
+#else
 	return _pure_syscall(__NR_rename,oldpath,newpath);
+#endif
 }
 
 int chdir(const char *path) {
@@ -383,7 +472,14 @@ int fchdir(int fd) {
 }
 
 int utimes(const char* pathname,const struct timeval tv[2]){
+#if defined(__NR_utimensat) && ! defined(__NR_utimes)
+	struct timespec ts[2] = {
+		{tv[0].tv_sec, tv[0].tv_usec * 1000},
+		{tv[1].tv_sec, tv[1].tv_usec * 1000}};
+	return _pure_syscall(__NR_utimensat, AT_FDCWD, pathname, ts, 0);
+#else
 	return _pure_syscall(__NR_utimes,pathname,tv);
+#endif
 }
 
 int utime(const char* pathname,const struct utimbuf *buf){
@@ -423,7 +519,7 @@ ssize_t pread(int fs,void* buf, size_t count, __off_t offset){
 }
 #endif
 
-#ifdef __NR_pwrite64 
+#ifdef __NR_pwrite64
 ssize_t pwrite64(int fs,const void* buf, size_t count, __off64_t offset){
 	return _pure_syscall(__NR_pwrite64,fs,buf,count,
 #if defined(__powerpc__) || defined(__arm__)
@@ -463,7 +559,7 @@ ssize_t preadv64(int fs,const struct iovec *iov, int iovcnt, __off64_t offset){
 		rv=totalsize;
 		for (i=0,scan=buf; i<iovcnt && totalsize>0; i++,
 				scan+=iov[i].iov_len,
-				totalsize-=iov[i].iov_len) 
+				totalsize-=iov[i].iov_len)
 			memcpy(iov[i].iov_base, scan, iov[i].iov_len);
 		free(buf);
 	}
@@ -480,7 +576,7 @@ ssize_t pwritev64(int fs,const struct iovec *iov, int iovcnt, __off64_t offset){
 	ssize_t rv=_pure_syscall(__NR_pwritev,fs,iov,iovcnt,
 #ifdef __NR_pwrite64
 #if defined(__powerpc__) || defined(__arm__)
-			0,    
+			0,
 #endif
 			__LONG_LONG_PAIR( (__off_t)(offset>>32),(__off_t)(offset&0xffffffff))
 #else
@@ -511,8 +607,15 @@ ssize_t pwritev(int fs,const struct iovec *iov, int iovcnt, __off_t offset){
 }
 #endif
 
+/* getdents has no libc wrapper.
+	 libc uses getdents64 only on 64 bit archs */
 int getdents(int fd, void *dirp, unsigned int count){
+#ifdef __NR_getdents
 	return _pure_syscall(__NR_getdents, fd, dirp, count);
+#else
+	errno = ENOSYS;
+	return -1;
+#endif
 }
 
 __ssize_t getdents64(int fd, void *dirp, size_t count){
@@ -766,7 +869,11 @@ int setresgid(gid_t rgid, gid_t egid, gid_t sgid){
 }
 
 int pipe(int filedes[2]) {
+#if defined(__NR_pipe2) && ! defined(__NR_pipe)
+	return _pure_syscall(__NR_pipe2,filedes, 0);
+#else
 	return _pure_syscall(__NR_pipe,filedes);
+#endif
 }
 
 #ifdef __NR_pipe2
@@ -801,7 +908,23 @@ void _Exit(int status){
 }
 
 pid_t fork(void){
+#if defined(__ia64__)
+	int child_tid;
+	if (_pure_syscall(__NR_clone2,
+				CLONE_CHILD_SETTID | CLONE_CHILD_CLEARTID | SIGCHLD,
+				NULL, 0, NULL, &child_tid, NULL) < 0)
+		return -1;
+	else
+		return child_tid;
+#elif defined(__aarch64__)
+	int child_tid;
+	if (_pure_syscall(__NR_clone, NULL, CLONE_CHILD_CLEARTID|CLONE_CHILD_SETTID|SIGCHLD, &child_tid) < 0)
+		return -1;
+	else
+		return child_tid;
+#else
 	return _pure_syscall(__NR_fork);
+#endif
 }
 
 pid_t vfork(void){
@@ -809,7 +932,7 @@ pid_t vfork(void){
 	 * proper system sharing mechanisms are  implemented.   Users  should  not
 	 * depend  on  the memory sharing semantics of vfork() as it will, in that
 	 * case, be made synonymous to fork(2)." */
-	return _pure_syscall(__NR_fork);
+	return fork();
 }
 
 int stime(const time_t *t){
@@ -839,11 +962,15 @@ long int ptrace (enum __ptrace_request request, ...){
 }
 
 int nice(int inc){
-#if ! defined(__x86_64__)
-	return _pure_syscall(__NR_nice,inc);
-#else
+#if defined(__x86_64__) || defined(__ia64__) || \
+	defined(__alpha__) || defined(__s390x__) || \
+	(defined(__mips__) && defined(__LP64__)) || \
+	defined(__aarch64__) || \
+	defined(__riscv__)
 	int nice = _pure_syscall(__NR_getpriority,PRIO_PROCESS,0);
 	return _pure_syscall(__NR_setpriority,PRIO_PROCESS,0,nice + inc);
+#else
+	return _pure_syscall(__NR_nice,inc);
 #endif
 }
 
@@ -857,7 +984,12 @@ clock_t times(struct tms *buf){
 
 struct ustat;
 int ustat(dev_t dev, struct ustat *ubuf){
+#ifdef __NR_ustat
 	return _pure_syscall(__NR_ustat,dev,ubuf);
+#else
+	errno = ENOSYS;
+	return -1;
+#endif
 }
 
 pid_t getsid(pid_t pid){
@@ -882,7 +1014,7 @@ int setrlimit(__rlimit_resource_t resource, const struct rlimit *rlim){
 	return prlimit(0,resource,rlim,NULL);
 }
 
-int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){ 
+int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){
 	return prlimit(0,resource,NULL,rlim);
 }
 #else
@@ -890,7 +1022,7 @@ int setrlimit(__rlimit_resource_t resource, const struct rlimit *rlim){
 	return _pure_syscall(__NR_setrlimit,resource,rlim);
 }
 
-int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){ 
+int getrlimit(__rlimit_resource_t resource, struct rlimit *rlim){
 	return _pure_syscall(__NR_getrlimit,resource,rlim);
 }
 #endif
@@ -929,21 +1061,24 @@ int getgroups(int size, gid_t list[]){
 	return _pure_syscall(__NR_getgroups,size,list);
 }
 
+#ifdef __NR_faccessat
+static int _is_group_member(gid_t gid) {
+  int len = getgroups(0, NULL);
+  gid_t list[len];
+  int i;
+  len = getgroups(len, list);
+  for (i = 0; i < len; i++) {
+    if (gid == list[i])
+      return 1;
+  }
+  return 0;
+}
+#endif
+
 int setgroups(size_t size, const gid_t *list){
 	return _pure_syscall(__NR_setgroups,size,list);
 }
 
-#if defined(__x86_64__)
-#define __NR__newselect __NR_select
-#endif
-int select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout){
-	return _pure_syscall(__NR__newselect,n,readfds,writefds,exceptfds,timeout);
-}
-
-int poll(struct pollfd *ufds, nfds_t nfds, int timeout){
-	return _pure_syscall(__NR_poll,ufds,nfds,timeout);
-}
-
 int pselect(int nfds, fd_set *readfds, fd_set *writefds,
 		fd_set *exceptfds, const struct timespec *timeout,
 		const sigset_t *sigmask) {
@@ -956,6 +1091,36 @@ int ppoll(struct pollfd *fds, nfds_t nfds,
 }
 
 #ifdef __NR_epoll_create1
+int select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout){
+#if defined(__x86_64__) || defined(__s390x__) || \
+	defined(__alpha__) || defined(__ia64__) || \
+	defined(__riscv__)
+	return _pure_syscall(__NR_select,n,readfds,writefds,exceptfds,timeout);
+#elif defined(__aarch64__)
+	if (timeout == NULL)
+		return pselect(n,readfds,writefds,exceptfds,NULL,NULL);
+	else {
+		struct timespec ts = {timeout->tv_sec, timeout->tv_usec * 1000};
+		return pselect(n,readfds,writefds,exceptfds,&ts,NULL);
+	}
+#else
+	return _pure_syscall(__NR__newselect,n,readfds,writefds,exceptfds,timeout);
+#endif
+}
+
+int poll(struct pollfd *ufds, nfds_t nfds, int timeout){
+#if defined(__NR_poll)
+	return _pure_syscall(__NR_poll,ufds,nfds,timeout);
+#else
+	if (timeout < 0)
+		return  _pure_syscall(__NR_ppoll,ufds,nfds, NULL, NULL);
+	else {
+		struct timespec ts = {timeout, 0};
+		return  _pure_syscall(__NR_ppoll,ufds,nfds, &ts, NULL);
+	}
+#endif
+}
+
 int epoll_create1(int flags) {
 	return _pure_syscall(__NR_epoll_create1, flags);
 }
@@ -1028,16 +1193,18 @@ int fstatfs(int fd, struct statfs *buf){
 	return _pure_syscall(__NR_fstatfs,fd,buf);
 }
 
-#if __WORDSIZE == 32
+#ifdef __NR_statfs64
 /* LIBC add an extra arg: the buf size */
 int statfs64(const char *path, struct statfs64 *buf){
 	return _pure_syscall(__NR_statfs64,path,sizeof(struct statfs64), buf);
 }
+#endif 
 
+#ifdef __NR_fstatfs64
 int fstatfs64(int fd, struct statfs64 *buf){
 	return _pure_syscall(__NR_fstatfs64,fd,sizeof(struct statfs64), buf);
 }
-#endif 
+#endif
 
 int getitimer(__itimer_which_t which, struct itimerval *value){
 	return _pure_syscall(__NR_getitimer,which,value);
@@ -1087,6 +1254,7 @@ int adjtimex(struct timex *buf){
 	return _pure_syscall(__NR_adjtimex,buf);
 }
 
+#ifdef __NR_sysfs
 int sysfs(int option,...){
 	switch (option) {
 		case 1: {
@@ -1114,6 +1282,7 @@ int sysfs(int option,...){
 						 return -1;
 	}
 }
+#endif
 
 int setfsuid(uid_t fsuid){
 	return _pure_syscall(__NR_setfsuid,fsuid);
@@ -1136,7 +1305,7 @@ char *getcwd(char *buf, size_t size){
 	if (size == 0 && buf==NULL) {
 		size=PATH_MAX;
 		buf=malloc(size);
-		if (buf==NULL) 
+		if (buf==NULL)
 			return NULL;
 		else {
 			rsize=_pure_syscall(__NR_getcwd,buf,size);
@@ -1150,7 +1319,7 @@ char *getcwd(char *buf, size_t size){
 		}
 	} else {
 		rsize=_pure_syscall(__NR_getcwd,buf,size);
-		if (rsize>=0) 
+		if (rsize>=0)
 			return buf;
 		else
 			return NULL;
@@ -1264,14 +1433,14 @@ int fstatvfs(int fd, struct statvfs *buf){
 void *mmap(void  *start, size_t length, int prot, int flags, int fd,
 		       off_t offset)
 {
-#if __WORDSIZE != 64
+#if defined(__NR_mmap2)
 		return (void *) _pure_syscall(__NR_mmap2,start,length,prot,flags,fd,offset>> _pageshift());
 #else
 		return (void *) _pure_syscall(__NR_mmap,start,length,prot,flags,fd,offset);
 #endif
 }
 
-#if __WORDSIZE != 64
+#if defined(__NR_mmap2)
 void *mmap2(void  *start, size_t length, int prot, int flags, int fd,
 		       off_t pgoffset)
 {
@@ -1377,12 +1546,6 @@ int futimesat(int dirfd, const char *pathname, const struct timeval times[2]) {
 }
 #endif
 
-#ifdef __NR_fstatat64
-#define __NR_FSTATAT64 __NR_fstatat64
-#endif
-#ifdef __NR_newfstatat
-#define __NR_FSTATAT64 __NR_newfstatat
-#endif
 #ifdef __NR_FSTATAT64
 int fstatat(int dirfd, const char *pathname, struct stat *buf, int flags) {
 	return _pure_syscall(__NR_FSTATAT64,dirfd,pathname,buf,flags);
@@ -1422,6 +1585,12 @@ int renameat(int olddirfd, const char *oldpath,int newdirfd, const char *newpath
 }
 #endif
 
+#ifdef __NR_renameat2
+int renameat2(int olddirfd, const char *oldpath,int newdirfd, const char *newpath, unsigned int flags){
+	return _pure_syscall(__NR_renameat2,olddirfd,oldpath,newdirfd,newpath,flags);
+}
+#endif
+
 #ifdef __NR_linkat
 int linkat(int olddirfd, const char *oldpath,int newdirfd, const char *newpath, int flags){
 	return _pure_syscall(__NR_linkat,olddirfd,oldpath,newdirfd,newpath,flags);
@@ -1447,8 +1616,61 @@ int fchmodat(int dirfd, const char *pathname, mode_t mode, int flags){
 #endif
 
 #ifdef __NR_faccessat
+//The raw faccessat() system call takes only the first  three  arguments
+// this function has been inspired by glibc: sysdeps/unix/sysv/linux/faccessat.c
 int faccessat(int dirfd, const char *pathname, int mode, int flags){
-	return _pure_syscall(__NR_faccessat,dirfd,pathname,mode,flags);
+	if (flags & ~(AT_EACCESS | AT_SYMLINK_NOFOLLOW)) {
+    errno = EINVAL;
+    return -1;
+  }
+
+	if (flags == 0)
+		return _pure_syscall(__NR_faccessat,dirfd,pathname,mode);
+	else {
+		struct stat stats;
+		if (fstatat(dirfd, pathname, &stats, flags & AT_SYMLINK_NOFOLLOW) < 0)
+			return -1;
+		mode &= (R_OK | W_OK | X_OK);
+		if (mode == F_OK)
+			return 0;
+		uid_t uid = (flags & AT_EACCESS) ? geteuid() : getuid();
+		if (uid == 0) { // it is root
+			if ((mode & X_OK) == 0) // RW are always allowed
+				return 0;
+			// X OK is X is okay for someone
+			if (stats.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
+				return 0;
+		}
+
+		int granted;
+		if (uid == stats.st_uid)
+			// user permissions
+			granted = (int) ((stats.st_mode >> 6) & mode);
+		else {
+			gid_t gid = (flags & AT_EACCESS) ? getegid() : getgid();
+			if (stats.st_gid == gid || _is_group_member(stats.st_gid))
+				// group permissions
+				granted = (int) ((stats.st_mode >> 3) & mode);
+			else
+				// other permissions
+				granted = stats.st_mode & mode;
+		}
+		if (granted == mode)
+			return 0;
+
+		errno = EACCES;
+		return -1;
+	}
+}
+
+int euidaccess(const char *pathname, int mode){
+	return faccessat(AT_FDCWD,pathname,mode,AT_EACCESS);
+}
+int eaccess(const char *pathname, int mode){
+	return euidaccess(pathname,mode);
+}
+int __euidaccess(const char *pathname, int mode){
+	return euidaccess(pathname,mode);
 }
 #endif
 
@@ -1475,7 +1697,7 @@ long int syscall(long int n,...)
 	arg4=va_arg(ap, long int);
 	arg5=va_arg(ap, long int);
 	va_end(ap);
-	/* in case of pre-init call: emergency initialization */ 
+	/* in case of pre-init call: emergency initialization */
 	if (__builtin_expect(_pure_native_syscall == NULL,0))
 		init();
 	if (__builtin_expect(_pure_syscall == syscall,0))
